@@ -11,7 +11,28 @@ import dask.array as da
 import warnings
 import struct
 import threading
+import inspect
 
+_POSITION_AXIS = "position"
+_ROW_AXIS = "row"
+_COLUMN_AXIS = "column"
+_Z_AXIS = "z"
+_TIME_AXIS = "time"
+_CHANNEL_AXIS = "channel"
+
+_AXIS_ORDER = {_ROW_AXIS: 7,
+               _COLUMN_AXIS: 6,
+               _POSITION_AXIS: 5, 
+               _TIME_AXIS: 4, 
+               _CHANNEL_AXIS:3, 
+               _Z_AXIS:2}
+
+def _get_axis_order_key(dict_item):
+    axis_name = dict_item[0]
+    if axis_name in _AXIS_ORDER.keys():
+        return _AXIS_ORDER[axis_name]
+    else:
+        return 3  # stack next to channel axes
 
 class _SingleNDTiffReader:
     """
@@ -21,9 +42,13 @@ class _SingleNDTiffReader:
 
     # file format constants
     SUMMARY_MD_HEADER = 2355492
-    EIGHT_BIT = 0
-    SIXTEEN_BIT = 1
+    EIGHT_BIT_MONOCHROME = 0
+    SIXTEEN_BIT_MONOCHROME = 1
     EIGHT_BIT_RGB = 2
+    TEN_BIT_MONOCHROME = 3
+    TWELVE_BIT_MONOCHROME = 4
+    FOURTEEN_BIT_MONOCHROME = 5
+
     UNCOMPRESSED = 0
 
     def __init__(self, tiff_path):
@@ -67,8 +92,8 @@ class _SingleNDTiffReader:
 
         # read custom stuff: header, summary md
         # int.from_bytes(self.mmap_file[24:28], sys.byteorder) # should be equal to 483729 starting in version 1
-        self._major_version = int.from_bytes(self._read(12, 16), sys.byteorder)
-        self._minor_version = int.from_bytes(self._read(16, 20), sys.byteorder)
+        self.major_version = int.from_bytes(self._read(12, 16), sys.byteorder)
+        self.minor_version = int.from_bytes(self._read(16, 20), sys.byteorder)
 
         summary_md_header, summary_md_length = np.frombuffer(self._read(20, 28), dtype=np.uint32)
         if summary_md_header != self.SUMMARY_MD_HEADER:
@@ -95,10 +120,13 @@ class _SingleNDTiffReader:
         if index["pixel_type"] == self.EIGHT_BIT_RGB:
             bytes_per_pixel = 3
             dtype = np.uint8
-        elif index["pixel_type"] == self.EIGHT_BIT:
+        elif index["pixel_type"] == self.EIGHT_BIT_MONOCHROME:
             bytes_per_pixel = 1
             dtype = np.uint8
-        elif index["pixel_type"] == self.SIXTEEN_BIT:
+        elif index["pixel_type"] == self.SIXTEEN_BIT_MONOCHROME or \
+                index["pixel_type"] == self.TEN_BIT_MONOCHROME or \
+                index["pixel_type"] == self.TWELVE_BIT_MONOCHROME or \
+                index["pixel_type"] == self.FOURTEEN_BIT_MONOCHROME:
             bytes_per_pixel = 2
             dtype = np.uint16
         else:
@@ -146,7 +174,7 @@ class NDTiffDataset():
         self._full_resolution = False
         self._lock = threading.RLock()
         if remote_storage_monitor is not None:
-            # this dataset is a view of an active acquisition. The storage exists on the java side
+            # this dataset is a view of an active acquisition. Image data is being written by code on the Java side
             self._new_image_arrived = False # used by custom (e.g. napari) viewer to check for updates. Will be reset to false by them
             self._remote_storage_monitor = remote_storage_monitor
             self.summary_metadata = self._remote_storage_monitor.get_summary_metadata()
@@ -156,6 +184,7 @@ class NDTiffDataset():
                 self.path = dataset_path #Overriden by pyramid storage class creating this
             self.path += "" if self.path[-1] == os.sep else os.sep
             self.axes = {}
+            self.axes_types = {}
             self.index = {}
             self._readers_by_filename = {}
             return
@@ -168,6 +197,8 @@ class NDTiffDataset():
             os.path.join(self.path, tiff) for tiff in os.listdir(self.path) if tiff.endswith(".tif")
         ]
         self._readers_by_filename = {}
+        self.summary_metadata = {}
+        self.major_version, self.minor_version = (0, 0)
         # Count how many files need to be opened
         num_tiffs = 0
         count = 0
@@ -178,8 +209,13 @@ class NDTiffDataset():
         for tiff in tiff_names:
             print("\rOpening file {} of {}...".format(count + 1, num_tiffs), end="")
             count += 1
-            self._readers_by_filename[tiff.split(os.sep)[-1]] = _SingleNDTiffReader(tiff)
-        self.summary_metadata = list(self._readers_by_filename.values())[0].summary_md
+            new_reader = _SingleNDTiffReader(tiff)
+            self._readers_by_filename[tiff.split(os.sep)[-1]] = new_reader
+            # Should be the same on every file so resetting them is fine
+            self.major_version, self.minor_version = new_reader.major_version, new_reader.minor_version
+
+        if len(self._readers_by_filename) > 0:
+            self.summary_metadata = list(self._readers_by_filename.values())[0].summary_md
 
 
         self.overlap = (
@@ -192,30 +228,45 @@ class NDTiffDataset():
         )
 
         self.axes = {}
+        self.axes_types = {}
         for axes_combo in self.index.keys():
             for axis, position in axes_combo:
                 if axis not in self.axes.keys():
                     self.axes[axis] = []
+                    self.axes_types[axis] = type(position)
                 self.axes[axis].append(position)
                 # get sorted unique elements
                 self.axes[axis] = sorted(list(set(self.axes[axis])))
+        # Sort axes according to _AXIS_ORDER
+        self.axes = dict(sorted(self.axes.items(), key=_get_axis_order_key, reverse=True))
 
         # figure out the mapping of channel name to position by reading image metadata
-        print("\rReading channel names...", end="")
-        self._read_channel_names()
-        print("\rFinished reading channel names", end="")
-
+        self._channels = {}
+        self._parse_string_axes()
 
         # get information about image width and height, assuming that they are consistent for whole dataset
-        # (which is not neccessarily true but convenient when it is)
-        with self._lock:
-            first_index = list(self.index.values())[0]
-        self._parse_first_index(first_index)
+        # (which is not necessarily true but convenient when it is)
+        self.bytes_per_pixel = 1
+        self.dtype = np.uint8
+        self.image_width, self.image_height = (0, 0)
+        if len(self.index) > 0:
+            with self._lock:
+                first_index = list(self.index.values())[0]
+            self._parse_first_index(first_index)
 
         print("\rDataset opened                ")
 
+    def __enter__(self):
+        return self
+
+    def __exit__(self, exc_type, exc_val, exc_tb):
+        self.close()
+    
     def get_channel_names(self):
-        return list(self.channels.keys())
+        """
+        :return: list of channel names (strings)
+        """
+        return list(self._channels.keys())
 
     def has_new_image(self):
         """
@@ -231,11 +282,10 @@ class NDTiffDataset():
 
     def has_image(
         self,
-        channel=0,
+        channel=None,
         z=None,
         time=None,
         position=None,
-        channel_name=None,
         row=None,
         col=None,
         **kwargs
@@ -244,7 +294,7 @@ class NDTiffDataset():
 
         Parameters
         ----------
-        channel : int
+        channel : int or str
             index of the channel, if applicable (Default value = None)
         z : int
             index of z slice, if applicable (Default value = None)
@@ -252,8 +302,6 @@ class NDTiffDataset():
             index of the time point, if applicable (Default value = None)
         position : int
             index of the XY position, if applicable (Default value = None)
-        channel_name : str
-            Name of the channel. Overrides channel index if supplied (Default value = None)
         row : int
             index of tile row for XY tiled datasets (Default value = None)
         col : int
@@ -267,19 +315,17 @@ class NDTiffDataset():
             indicating whether the dataset has an image matching the specifications
         """
         with self._lock:
-            return self._does_have_image(
-                _consolidate_axes(self.channels, channel, channel_name, z, position, time, row, col, kwargs)
-            )
+            return self._does_have_image(self._consolidate_axes(
+                channel, z, position, time, row, col, **kwargs))
 
     def read_image(
         self,
-        channel=0,
+        channel=None,
         z=None,
         time=None,
         position=None,
         row=None,
         col=None,
-        channel_name=None,
         memmapped=False,
         **kwargs
     ):
@@ -288,7 +334,7 @@ class NDTiffDataset():
 
         Parameters
         ----------
-        channel : int
+        channel : int or str
             index of the channel, if applicable (Default value = None)
         z : int
             index of z slice, if applicable (Default value = None)
@@ -296,8 +342,6 @@ class NDTiffDataset():
             index of the time point, if applicable (Default value = None)
         position : int
             index of the XY position, if applicable (Default value = None)
-        channel_name :
-            Name of the channel. Overrides channel index if supplied (Default value = None)
         row : int
             index of tile row for XY tiled datasets (Default value = None)
         col : int
@@ -314,19 +358,16 @@ class NDTiffDataset():
 
         """
         with self._lock:
-            axes = _consolidate_axes(self.channels,
-                channel, channel_name, z, position, time, row, col, kwargs
-            )
+            axes = self._consolidate_axes(channel, z, position, time, row, col, **kwargs )
 
             return self._do_read_image(axes, memmapped)
 
     def read_metadata(
         self,
-        channel=0,
+        channel=None,
         z=None,
         time=None,
         position=None,
-        channel_name=None,
         row=None,
         col=None,
         **kwargs
@@ -336,7 +377,7 @@ class NDTiffDataset():
 
         Parameters
         ----------
-        channel : int
+        channel : int or str
             index of the channel, if applicable (Default value = None)
         z : int
             index of z slice, if applicable (Default value = None)
@@ -344,8 +385,6 @@ class NDTiffDataset():
             index of the time point, if applicable (Default value = None)
         position : int
             index of the XY position, if applicable (Default value = None)
-        channel_name :
-            Name of the channel. Overrides channel index if supplied (Default value = None)
         row : int
             index of tile row for XY tiled datasets (Default value = None)
         col : int
@@ -360,8 +399,8 @@ class NDTiffDataset():
 
         """
         with self._lock:
-            axes = _consolidate_axes(self.channels,
-                channel, channel_name, z, position, time, row, col, kwargs
+            axes = self._consolidate_axes(
+                channel, z, position, time, row, col, **kwargs
             )
 
             return self._do_read_metadata(axes)
@@ -379,48 +418,92 @@ class NDTiffDataset():
     def _add_index_entry(self, data):
         """
         Add entry for a image that has been recieved and is now on disk
+        This is used to get access to a dataset that is currently being
+        written by java side
         """
         with self._lock:
             _, axes, index_entry = self._read_single_index_entry(data, self.index)
 
             if index_entry["filename"] not in self._readers_by_filename:
-                self._readers_by_filename[index_entry["filename"]] = _SingleNDTiffReader(
-                    self.path + index_entry["filename"]
-                )
+                new_reader = _SingleNDTiffReader(self.path + index_entry["filename"] )
+                self._readers_by_filename[index_entry["filename"]] = new_reader
+                # Should be the same on every file so resetting them is fine
+                self.major_version, self.minor_version = new_reader.major_version, new_reader.minor_version
+
 
             # update the axes that have been seen
             for axis_name in axes.keys():
                 if axis_name not in self.axes.keys():
                     self.axes[axis_name] = []
+                    self.axes_types[axis_name] = type(axes[axis_name])
                 self.axes[axis_name].append(axes[axis_name])
                 self.axes[axis_name] = sorted(list(set(self.axes[axis_name])))
 
             # update the map of channel names to channel indices
-            self._read_channel_names()
+            self._parse_string_axes()
 
         if not hasattr(self, 'image_width'):
             self._parse_first_index(index_entry)
 
         return axes
 
-    def _read_channel_names(self):
-        if 'ChNames' in self.summary_metadata:
-            # It was created by a MM MDA/Clojure acquistiion engine
-            self.channels = {name: i for i, name in enumerate(self.summary_metadata['ChNames'])}
+    def _consolidate_axes(self, channel: int or str, z: int, position: int,
+                          time: int, row: int, col: int, **kwargs):
+        """
+        Pack axes into a convenient format
+        """
+        if ('channel_name' in kwargs):
+            warnings.warn('channel_name is deprecated, use "channel" instead')
+            channel = kwargs['channel_name']
+            del kwargs['channel_name']
+
+        axis_positions = {'channel': channel, 'z': z, 'position': position,
+                    'time': time, 'row': row, 'col': col, **kwargs}
+        # ignore ones that are None
+        axis_positions = {n: axis_positions[n] for n in axis_positions.keys() if axis_positions[n] is not None}
+        for axis_name in axis_positions.keys():
+            # convert any string-valued axes passed as ints into strings
+            if self.axes_types[axis_name] == str and type(axis_positions[axis_name]) == int:
+                axis_positions[axis_name] = self._string_axes_values[axis_name][axis_positions[axis_name]]
+
+        return axis_positions
+
+    def _parse_string_axes(self):
+        """
+        As of NDTiff 3.2, axes are allowed to take string values: e.g. {'channel': 'DAPI'}
+        This is allowed on any axis. This function returns a tuple of possible values along
+        the string axis in order to be able to interconvert integer values and string values.
+        """
+        # iterate through the key_combos for each image
+        if self.major_version >= 3 and self.minor_version >= 2:
+            self._string_axes_values = {axis_name: [] for axis_name in self.axes_types.keys()
+                                        if self.axes_types[axis_name] is str}
+            for key in self.index.keys():
+                for axis_name, position in key:
+                    if axis_name in self._string_axes_values.keys() and \
+                            position not in self._string_axes_values[axis_name]:
+                        self._string_axes_values[axis_name].append(position)
+            if _CHANNEL_AXIS in self._string_axes_values:
+                self._channels = {name: self._string_axes_values[_CHANNEL_AXIS].index(name)
+                                  for name in self._string_axes_values[_CHANNEL_AXIS]}
         else:
-            # AcqEngJ
-            self.channels = {}
-            if _CHANNEL_AXIS in self.axes.keys():
-                for key in self.index.keys():
-                    axes = {axis: position for axis, position in key}
-                    if (
-                        _CHANNEL_AXIS in axes.keys()
-                        and axes[_CHANNEL_AXIS] not in self.channels.values()
-                    ):
-                        channel_name = self.read_metadata(**axes)["Channel"]
-                        self.channels[channel_name] = axes[_CHANNEL_AXIS]
-                    if len(self.channels.values()) == len(self.axes[_CHANNEL_AXIS]):
-                        break
+            # before string-valued axes were allowed in NDTiff 3.1
+            if 'ChNames' in self.summary_metadata:
+                # It was created by a MM MDA/Clojure acquistiion engine
+                self._channels = {name: i for i, name in enumerate(self.summary_metadata['ChNames'])}
+            else:
+                # AcqEngJ
+                if _CHANNEL_AXIS in self.axes.keys():
+                    for key in self.index.keys():
+                        axes = {axis: position for axis, position in key}
+                        if (
+                            _CHANNEL_AXIS in axes.keys()
+                            and axes[_CHANNEL_AXIS] not in self._channels.values()
+                        ):
+                            channel_name = self.read_metadata(**axes)["Channel"]
+                            self._channels[channel_name] = axes[_CHANNEL_AXIS]
+                        if len(self._channels.values()) == len(self.axes[_CHANNEL_AXIS]):
+                            break
 
 
     def _parse_first_index(self, first_index):
@@ -430,10 +513,13 @@ class NDTiffDataset():
         if first_index["pixel_type"] == _SingleNDTiffReader.EIGHT_BIT_RGB:
             self.bytes_per_pixel = 3
             self.dtype = np.uint8
-        elif first_index["pixel_type"] == _SingleNDTiffReader.EIGHT_BIT:
+        elif first_index["pixel_type"] == _SingleNDTiffReader.EIGHT_BIT_MONOCHROME:
             self.bytes_per_pixel = 1
             self.dtype = np.uint8
-        elif first_index["pixel_type"] == _SingleNDTiffReader.SIXTEEN_BIT:
+        elif first_index["pixel_type"] == _SingleNDTiffReader.SIXTEEN_BIT_MONOCHROME or \
+                first_index["pixel_type"] == _SingleNDTiffReader.FOURTEEN_BIT_MONOCHROME or \
+                first_index["pixel_type"] == _SingleNDTiffReader.TWELVE_BIT_MONOCHROME or \
+                first_index["pixel_type"] == _SingleNDTiffReader.TEN_BIT_MONOCHROME:
             self.bytes_per_pixel = 2
             self.dtype = np.uint16
 
@@ -584,8 +670,8 @@ class NDTiffDataset():
         Parameters
         ----------
         axes : list
-            list of axes names over which to iterate and merge into a stacked array. If None, all axes will be used.
-            The order of axes supplied in this list will be the order of the axes of the returned dask array
+            list of axes names over which to iterate and merge into a stacked array. The order of axes supplied in this 
+            list will be the order of the axes of the returned dask array. If None, all axes will be used in PTCZYX order.
         stitched : bool
             If true and tiles were acquired in a grid, lay out adjacent tiles next to one another (Default value = False)
         **kwargs :
@@ -630,7 +716,6 @@ class NDTiffDataset():
             axes = {key: block_id[i] for i, key in enumerate(axes_to_stack.keys())}
             if stitched:
                 # Combine all rows and cols into one stitched image
-                self.half_overlap = (self.overlap[0] // 2, self.overlap[1] // 2)
                 # get spatial layout of position indices
                 row_values = np.array(list(self.axes["row"]))
                 column_values = np.array(list(self.axes["column"]))
@@ -649,11 +734,10 @@ class NDTiffDataset():
                             tile = self.read_image(**axes, **axes_to_slice, row=row, column=column, memmapped=True)
                             # remove half of the overlap around each tile so that that image stitches correctly
                             # only need this for full resoution because downsampled ones already have the edges removed
-                            if self.half_overlap[0] != 0 and self._full_resolution:
-                                tile = tile[
-                                       self.half_overlap[0]: -self.half_overlap[0],
-                                       self.half_overlap[1]: -self.half_overlap[1],
-                                       ]
+                            if np.any(self.overlap[0] > 0) and self._full_resolution:
+                                min_index = np.floor(self.overlap / 2).astype(np.int_)
+                                max_index = np.ceil(self.overlap / 2).astype(np.int_)
+                                tile = tile[min_index[0]:-max_index[0], min_index[1]:-max_index[1]]
                             blocks[-1].append(tile)
 
                 if rgb:
@@ -765,7 +849,6 @@ class NDTiffPyramidDataset():
                 )
 
                 self.axes = res_level.axes
-                self._channel_names = res_level.channels
                 self.bytes_per_pixel = res_level.bytes_per_pixel
                 self.dtype = res_level.dtype
                 self.image_width = res_level.image_width
@@ -848,8 +931,8 @@ class NDTiffPyramidDataset():
                 else:
                     image = self.res_levels[res_level].as_array(axes=axes, stitched=stitched, **kwargs)
                     # crop away zero padding that extends pass where data is collected
-                    res_level_pixel_extent_min = (pixel_extent_min / 2 ** res_level).astype(np.int)
-                    res_level_pixel_extent_max = (pixel_extent_max / 2 ** res_level).astype(np.int)
+                    res_level_pixel_extent_min = (pixel_extent_min / 2 ** res_level).astype(np.int_)
+                    res_level_pixel_extent_max = (pixel_extent_max / 2 ** res_level).astype(np.int_)
                     if np.min(np.stack([res_level_pixel_extent_max - res_level_pixel_extent_min,
                                         res_level_pixel_extent_max - res_level_pixel_extent_min])) < 16:
                         # Not worth it to use ones this small
@@ -872,7 +955,7 @@ class NDTiffPyramidDataset():
 
     def has_image(
         self,
-        channel=0,
+        channel=None,
         z=None,
         time=None,
         position=None,
@@ -925,7 +1008,7 @@ class NDTiffPyramidDataset():
 
     def read_image(
         self,
-        channel=0,
+        channel=None,
         z=None,
         time=None,
         position=None,
@@ -982,7 +1065,7 @@ class NDTiffPyramidDataset():
 
     def read_metadata(
         self,
-        channel=0,
+        channel=None,
         z=None,
         time=None,
         position=None,
@@ -1024,7 +1107,7 @@ class NDTiffPyramidDataset():
         """
         with self._lock:
             return self.res_levels[resolution_level].read_metadata(
-                    channel=0,
+                    channel=channel,
                     z=z,
                     time=time,
                     position=position,
@@ -1046,38 +1129,3 @@ class NDTiffPyramidDataset():
         with self._lock:
             for res_level in self.res_levels:
                 res_level.close()
-
-
-_POSITION_AXIS = "position"
-_ROW_AXIS = "row"
-_COLUMN_AXIS = "column"
-_Z_AXIS = "z"
-_TIME_AXIS = "time"
-_CHANNEL_AXIS = "channel"
-
-def _consolidate_axes(channels: dict, channel: int, channel_name: str, z, position, time, row, col, kwargs):
-    """
-    Get all axis names in a consistent format
-    """
-    axes = {}
-    if channel is not None:
-        axes[_CHANNEL_AXIS] = channel
-    if channel_name is not None:
-        if channel_name in channels.keys():
-            axes[_CHANNEL_AXIS] = channels[channel_name]
-        else:
-            raise Exception(f'{channel_name} is not a valid channel name.'
-                            f'Valid channel names are: {list(channels.keys())}')
-    if z is not None:
-        axes[_Z_AXIS] = z
-    if position is not None:
-        axes[_POSITION_AXIS] = position
-    if time is not None:
-        axes[_TIME_AXIS] = time
-    if row is not None:
-        axes[_ROW_AXIS] = row
-    if col is not None:
-        axes[_COLUMN_AXIS] = col
-    for other_axis_name in kwargs.keys():
-        axes[other_axis_name] = kwargs[other_axis_name]
-    return axes
