@@ -75,6 +75,12 @@ class _MultipageTiffReader:
         summary_md = json.loads(self._read(24, 24 + summary_md_length))
         return summary_md, first_ifd_offset
 
+    def __enter__(self):
+        return self
+
+    def __exit__(self, exc_type, exc_val, exc_tb):
+        self.close()
+    
     def _read(self, start, end):
         """
         convert to python ints
@@ -718,7 +724,7 @@ class NDTiff_v2_0():
 
     def close(self):
         with self._lock:
-            for res_level in self.res_levels:
+            for res_level in self.res_levels.values():
                 res_level.close()
 
     def get_channel_names(self):
